#ifndef DRVXMAP_H
#define DRVXMAP_H

#include <asynDriver.h>
#include <devLib.h>
#include <mca.h>
#include <drvMca.h>

#include <handel.h>
#include <handel_errors.h>
#include <md_generic.h>
#include <handel_constants.h>

#define XMAP_FIRST_MCA_CMD mcaData
#define XMAP_LAST_MCA_CMD  mcaElapsedCounts
#define XMAP_FIRST_XIA_CMD xia_peaking_time
#define XMAP_LAST_XIA_CMD  xia_buffer_overrun
#define XMAP_FIRST_CMD     XMAP_LAST_XIA_CMD + 1
#define XMAP_LAST_CMD      XMAP_TOTAL_CMD - 1


/* Enumerate all available ASYN commands for this driver */
typedef enum{
	xia_peaking_time = (XMAP_LAST_MCA_CMD + 1),
	xia_dynamic_range,
	xia_trigger_threshold,
	xia_baseline_threshold,
	xia_energy_threshold,
	xia_calibration_energy,
	xia_adc_percent_rule,
	xia_mca_bin_width,
	xia_preamp_gain,
	xia_number_mca_channels,
	xia_detector_polarity,
	xia_reset_delay,
	xia_gap_time,
	xia_trigger_peaking_time,
	xia_trigger_gap_time,
	xia_baseline_average,
	xia_preset_type,
	xia_preset_value,
	xia_number_of_scas,
	xia_mapping_mode,
	xia_num_map_pixels,
	xia_num_map_pixels_per_buffer,
	xia_input_logic_polarity,
	xia_gate_master,
	xia_sync_master,
	xia_sync_count,
	xia_gate_ignore,
	xia_lbus_master,
	xia_pixel_advance_mode,
	xia_minimum_gap_time,
	xia_synchronous_run,
	xia_maxwidth,
	xia_preamp_type,
	xia_decay_time,
	xia_sca_n_lohi,
	xia_peak_sample_offset_n,
	xia_apply,
	xia_buffer_done,
	xia_mapping_pixel_next,
	xia_mca_length,
	xia_mca,
	xia_baseline_length,
	xia_baseline,
	xia_runtime,
	xia_realtime,
	xia_events_in_run,
	xia_trigger_livetime,
	xia_livetime,
	xia_input_count_rate,
	xia_output_count_rate,
	xia_sca_length,
/*	xia_sca,*/ /* currently not supporting float64Array */
	xia_run_active,
	xia_buffer_full_a,
	xia_buffer_full_b,
	xia_buffer_len,
	xia_buffer_a,
	xia_buffer_b,
	xia_current_pixel,
	xia_buffer_overrun,
/*	xia_module_statistics,
	xia_module_mca,   */

	xmap_mca_collect,		/* uint32 write - triggers a 4ch MCA collection */
	xmap_mca_stat,        /* uint32 write - triggers a 4ch MCA status read */
	xmap_mca_triggers,
	xmap_mca_preset_type,
	xmap_mca_preset_value,
	xmap_mca_acqstat,
	xmap_start_all,
	xmap_some_command,
	xmap_another_command,
	xmap_run,
	xmap_trace,
	xmap_enable_ch,
	xmap_enable_all,

	XMAP_TOTAL_CMD,		/* must always be last! */
	} xmap_cmd_enum;

#define XMAP_N_CMDS 13
#define XMAP_XIA_CMDS 59



/* bit masks of supported interface types */
#define XMAP_INT32        0x01
#define XMAP_UINT32       0x02
#define XMAP_FLOAT64      0x04
#define XMAP_INT32ARRAY   0x08
#define XMAP_NONE_TYPE    0x1F
#define XMAP_FLOAT64ARRAY 0x3F

#define XMAP_ALL -1
#define XMAP_NCHANS_MODULE 4
#define XMAP_MAX_MCA_BINS 16384
#define XMAP_MIN_MCA_BINS 256
#define XMAP_N_RUNTIME_STATS 7
#define XMAP_MCA_BIN_RES 256
#define TRUE 1
#define FALSE 0

/* preset types defined here. Note these currently map
 * directly the defines in handel_constants.h */
typedef enum xmap_preset_enum {
	XMAP_PRESET_NONE = 0,
	XMAP_PRESET_REAL,
	XMAP_PRESET_LIVE,
	XMAP_PRESET_EVENTS,
	XMAP_PRESET_TRIGGERS
	} xmap_preset_enum;

/* Supported modes on the XIA xMAP */
typedef enum{ XMAP_MODE_MCA = 0, XMAP_MODE_MAPPING, XMAP_MODE_SCA }xmap_mode_enum;
typedef enum{ XMAP_GET, XMAP_SET }xmap_dir_enum;
typedef enum{ XMAP_XIA_ACQ, XMAP_XIA_RUN, XMAP_XIA_BOARD, XMAP_MCA, XMAP_XMAP, XMAP_CMD_INVALID }xmap_support_t;
typedef enum{ 	XMAP_TRACE_ADC, XMAP_TRACE_BASELINE_HISTORY, XMAP_TRACE_TRIGGER_FILTER, XMAP_TRACE_BASELINE_FILTER,
				XMAP_TRACE_ENERGY_FILTER, XMAP_TRACE_BASELINE_SAMPLES, XMAP_TRACE_ENERGY_SAMPLES} xmap_tracemode_t;
typedef enum{ XMAP_START_MODULE, XMAP_START_ALL }xmap_start_enum;

typedef struct {
	xmap_cmd_enum command;
	char *commandString;
	char *xiaCmd;
	unsigned int interfaceType;
} xmap_command_t;

typedef struct xmap_channel_t
{
	epicsInt32 nBins;
	epicsInt32 acquiring;
	epicsInt32 *mcaChData;
	epicsInt32 *mcaModuleData;
	epicsInt32 enabled;
	epicsInt32 resume;

	/* module runtime stats */
	epicsFloat64 realtime;
	epicsFloat64 trigger_livetime;
	epicsFloat64 energy_livetime;
	epicsFloat64 triggers;
	epicsFloat64 events;
	epicsFloat64 input_count_rate;
	epicsFloat64 output_count_rate;
	epicsUInt32 run_active_status;
}xmap_channel_t;

/* internal driver structure */
typedef struct xmap_driver_t
{
	ELLNODE node;
	char * portName;
	char * initFile;
	xmap_start_enum start_mode;
	epicsUInt32 mode;
	epicsFloat64 tmpModuleStats[XMAP_NCHANS_MODULE * XMAP_N_RUNTIME_STATS];
	epicsUInt32 xch[XMAP_NCHANS_MODULE];	/* mapping of index/channels on the current module */
	epicsUInt32 moduleNo;

	xmap_channel_t *chan;

	/* xMAP in mapping/buffering fast acq mode */
	char current_xiabuf;

	epicsUInt32 resume;
	epicsInt32 acquiring;

	/* preset run conditions */
	epicsInt32 preset_type;
	epicsFloat64 preset_value;

	/* ADC Trace settings */
	epicsFloat64 trace_sampletime;
	epicsInt32* trace_large_buf;
	epicsInt32 trace_len;
	xmap_tracemode_t trace_mode;

	/* asyn bits */
	asynUser *localpasynUser;
	/* asyn interfaces to implement for this driver */
	asynInterface common;
	asynInterface int32;
	asynInterface float64;
	asynInterface int32Array;
	asynInterface uint32Digital;
	asynInterface drvUser;
	/* asyn interrupt (Inst I/O) interfaces to implement */
	void *int32InterruptPvt;
	void *float64InterruptPvt;
	void *int32ArrayInterruptPvt;
	void *uint32DigitalInterruptPvt;
} xmap_driver_t;

/* Configuration function prototype */
int xmapConfig(char * portName, int nModule, char* initfile);
void xmapClose(void);

/* local functions */
asynStatus xmap_action( asynUser *pasynUser, xmap_driver_t* pPvt, epicsInt32 reason, epicsInt32 value);
asynStatus xmap_status( asynUser *pasynUser, xmap_driver_t* pPvt, epicsInt32 reason, epicsInt32* value);
asynStatus xmap_parameter( asynUser *pasynUser, xmap_driver_t* pPvt, epicsInt32 reason, epicsInt32* value);
asynStatus xmap_data( asynUser *pasynUser, xmap_driver_t* pPvt, epicsInt32 reason, epicsInt32* data, size_t* len);

xmap_support_t xmap_get_support_type(epicsInt32 command);

/* local XIA/handel library functions */
asynStatus xmap_xia_acq_val( asynUser* pasynUser, epicsFloat64* value, xmap_dir_enum dir);
asynStatus xmap_xia_run_val(asynUser* pasynUser, void* value, epicsInt32* len, unsigned int interfaceType);
asynStatus xmap_xia_board_op(asynUser* pasynUser, char* buf);
asynStatus xmap_xia_checkError( asynUser* pasynUser, epicsInt32 xiastatus, char *xiacmd );
xmap_command_t xmap_xia_get_cmdreason( epicsInt32 reason );


/*local MCA xmap functions*/
asynStatus xmap_mca_int_write        (xmap_driver_t* pPvt, asynUser *pasynUser, epicsInt32 value);
asynStatus xmap_mca_int_read         (xmap_driver_t* pPvt, asynUser *pasynUser, epicsInt32 *value);
asynStatus xmap_mca_int_array_read   (xmap_driver_t* pPvt, asynUser *pasynUser, epicsInt32 *data, size_t maxChans, size_t *nactual);
asynStatus xmap_int_array_write      (xmap_driver_t* pPvt, asynUser *pasynUser, epicsInt32 *data, size_t maxChans);
asynStatus xmap_mca_float_write      (xmap_driver_t* pPvt, asynUser *pasynUser, epicsFloat64 value);
asynStatus xmap_mca_float_read       (xmap_driver_t* pPvt, asynUser *pasynUser, epicsFloat64 *value);

#endif
